#
# Licensed Materials - Property of IBM
# Rational ClearQuest
# (C) Copyright IBM Corp. 2007, 2014. All Rights Reserved.
#
# US Government Users Restricted Rights -
# Use, duplication or disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
#

# Load the ClearQuest API interface...
use CQPerlExt;

use strict;

sub Usage {
	print <<EOF;

This script is a used to find ClearQuest record locks and do lock timeout.

Usage: cqperl findrecordlocks.pl [-dbset <dbset_name>]
                                 [-database <user_database>]
                                 <clearquest_login>
                                 <clearquest_password>
                                 { <record_type_list> | -all }
                                 [-unlock]
                                 [-scan <interval>]
                                 [-timeout <interval>]
                                 [-times <times>]
                                 [-loglevel <error|info|verbose>]

Parameters:
  -dbset       A ClearQuest dbset name, the default dbset will be used if this
               option is not provided
  -database:   The user database to scan, this option is required if there is
               more than one user database in the dbset
  clearquest_login      Login name of the dbset
  clearquest_password   Password of the login name
  record_type_list      A list of record types separated by space or comma
  -all:        All record types in the schema
  -unlock:     Automatically release the locks if any timed-out locks are found
  -scan:       The interval to check locks in seconds
  -timeout:    Interval time in seconds for locks deemed as time-out
               (default=3600)
  -times:      The number of times to scan, default is unlimited
  -loglevel:   Specify what level of message should be printed out on the
               stdout (error, info, or verbose default='verbose')

It runs periodically to look for locked records on specified record types. If
-unlock option is not given, this utility only reports the record locks without
trying to release them. This utility can find locks on all record types or a
list of specified record types.

Example 1. Report ClearQuest dbset locked records in record type Defect:
cqperl findrecordlocks.pl -dbset RecordLocking admin "" Defect

Example 2. Monitor ClearQuest dbset locked records every 7200 seconds (2 hours)
in the user database udb, scan 12 times totally:
cqperl findrecordlocks.pl -dbset RecordLocking -database udb admin "" Defect -scan 7200 -times 12

Example 3. Report time-out locked records but not unlock with 2 hours scan
interval and 4 hours timeout interval in Defect and Requirement record types:
cqperl findrecordlocks.pl -dbset RecordLocking admin "" Defect,Requirement -scan 7200 -timeout 18000 -loglevel info

Example 4. Time-out Mode: Unlock abandoned locked records every 4 hours:
cqperl findrecordlocks.pl -dbset RecordLocking admin "" -all -unlock -scan 7200 -timeout 18000 -loglevel verbose

EOF
	exit 1;
}

# Define default values...
my(%ARG) = (
    DBSET     => '',        # 7.1 default dbset
    UNLOCK    => 0,         # false
    SCAN      => 0,         # Report only
    DATABASE  => '',        # User database name
    TIMEOUT   => 3600,      # 1 hour
    LOGLEVEL  => 'verbose',
    TIMES     => 0,
);

# See if they requested usage info...
if ((@ARGV == 0) || ($ARGV[0] =~ /^-[h\?]/i)) {
    Usage();
    exit;
}

# Parse the user's command-line parameters...
my $idx = 0;
my $arg = $ARGV[$idx++];

if ($arg =~ /-dbset/i) {
    $ARG{DBSET} = $ARGV[$idx++];
    $arg = $ARGV[$idx++];
}

if ($arg =~ /-database/i) {
    $ARG{DATABASE} = $ARGV[$idx++];
    $arg = $ARGV[$idx++];
}

$ARG{CQUSER} = $arg;
$ARG{CQPASSWORD} = $ARGV[$idx];

# Mask ClearQuest Password
$ARGV[$idx++] = '*';

$ARG{RECORDTYPE} = $ARGV[$idx++];

while ($idx < @ARGV) {
    $arg = $ARGV[$idx++];

    if ($arg =~ /^-unlock$/i) {
        $ARG{UNLOCK} = 1;
    }
    elsif ($arg =~ /^-(scan|timeout|times|loglevel)$/i) {
        # Keyword/value parameter; get the value...
        $ARG{uc $1} = $ARGV[$idx++];
    }
    elsif ($arg =~ /^-/) {
        die "ERROR: Unrecognized argument '$arg'.";
    }
    else {
        # Consider non-option argument as a record type name
        $ARG{RECORDTYPE} .= ','.$arg;
    }
}

# Verify that we have all required args...
my(@Errors);
push @Errors, "Missing parameter 'clearquest_login'" if (not $ARG{CQUSER});
push @Errors, "Missing parameter 'record_type_list'" if (not $ARG{RECORDTYPE});

if (not $ARG{SCAN} =~ /\d+$/) {
    push @Errors, "-scan must be a number";
}
if (not $ARG{TIMEOUT} =~ /\d+$/) {
    push @Errors, "-timeout must be a number";
}
if (not $ARG{LOGLEVEL} =~ /error|info|verbose/i) {
    push @Errors, "-loglevel must be 'error', 'info', or 'verbose'";
}

if (@Errors) {
    print join("\n", @Errors)."\n\n";
    Usage();
    exit 1;
}

sub UnlockRecord($$);
sub DisplayName($$);
sub Logs($);

my $scantimes = 0;
if ($ARG{TIMES} =~ /\d+$/) {
    $scantimes = $ARG{TIMES};
}

# Start a Rational ClearQuest session
my $cqSession = CQSession::Build();

# Get user database name
# "MASTR" is ClearQuest logical database name of the schema repository
if ($ARG{DATABASE} eq '') {
    my $databases = $cqSession->GetAccessibleDatabases("MASTR",
                                    $ARG{CQUSER}, $ARG{DBSET});
    if (($databases->Count()) == 1) {
        $ARG{DATABASE} = $databases->Item(0)->GetDatabaseName();
    }
    else {
        die 'There is more than one user database in dbset '
            .$ARG{DBSET}.', please specify intended user database using '
            .'-database option.';
    }
}

# Login the user database as specified ClearQuest user
$cqSession->UserLogon($ARG{CQUSER}, $ARG{CQPASSWORD}, $ARG{DATABASE},
                      $ARG{DBSET});

# Handle -all option
my $allTypes = 0;
if ($ARG{RECORDTYPE} =~ /-all/i) {
    my $entityDefNames = $cqSession->GetEntityDefNames();
    $ARG{RECORDTYPE} = join(',', sort(@$entityDefNames));
    $allTypes = 1;
}

# Echo arguments if verbose loglevel
if ($ARG{LOGLEVEL} =~ /verbose/i) {
    &Logs('findrecordlocks.pl '.join(' ', @ARGV));

    # A blank line with timestamp prefix
    &Logs('');
}

# Get non-system owned record types
my ($recordType, @recordTypes, $entityDef);
foreach $recordType (split(',+', $ARG{RECORDTYPE})) {
    eval { $entityDef = $cqSession->GetEntityDef($recordType); };
    if ($@) {
        &Logs($@);
    }
    elsif ($cqSession->GetEntityDef($recordType)->IsBuiltInSystemOwned()) {
        if (($allTypes == 0) && ($ARG{LOGLEVEL} =~ /info|verbose/i)) {
            &Logs("Warning: $recordType is a system-owned record type that "
                  ."cannot be locked and will not be scanned.");
        }
    }
    elsif (scalar(grep(/^$recordType$/i, @recordTypes)) == 0) {
        push(@recordTypes, $recordType);
    }
}

CQSession::Unbuild($cqSession);

# Sort the record type(s)
@recordTypes = sort(@recordTypes);

# Tell the user what we're going to do...
if ($ARG{LOGLEVEL} =~ /verbose/i) {
    &Logs('Looking for record locks on the following record types:');
    &Logs('  '.join(', ', @recordTypes));

    if ($ARG{SCAN} > 0) {
        &Logs("Scan interval is $ARG{SCAN} seconds");

        if ($ARG{TIMES} > 0) {
            &Logs("Scan times is $ARG{TIMES}");
        }

        if ($ARG{TIMEOUT} > 0) {
            &Logs("Timeout is $ARG{TIMEOUT} seconds");

            if ($ARG{UNLOCK} == 1) {
                &Logs("This script will release the abandoned locks");
            }
        }
    }

    &Logs('');
}

# Define id, version, locked_by, and lock version fields
my $dbidField = "dbid";
my $versionField = "version";
my $lockOwnerField = "locked_by";
my $lockVersionField = "lock_version";

# Set lock records information
my ($dbid, $version, $lockOwner, $lockVersion, $beginTime, $scanTime, $locks);

# Define lock records hashes
my ($lockKey, %lockRecords, %lastLockRecords, $timeoutLocks, $unlocks);

# Define query variables for loop below
my ($queryDef, $filterOp, @unlockValue, $resultset, $timeout, $logmsg);

# Get timeout count
if ($ARG{SCAN} > 0) {
    $timeout = $ARG{TIMEOUT} / $ARG{SCAN};
}

# Scan summary count
my ($total, $totaltimeout, $totalunlocks);

# Scan the record database frequently
while ($scantimes >= 0) {
    # Get begin time for exact interval scan
    $beginTime = time();

    # Clean up the lock records information
    %lockRecords = ();

    # Clean scan summary count
    $total = 0;
    $totaltimeout = 0;
    $totalunlocks = 0;

    # Start a Rational ClearQuest session
    $cqSession = CQSession::Build();

    # Login the user database as specified ClearQuest user
    $cqSession->UserLogon($ARG{CQUSER}, $ARG{CQPASSWORD}, $ARG{DATABASE},
                          $ARG{DBSET});

    foreach $recordType (@recordTypes) {
        # Build a query to look for lock records
        $queryDef = $cqSession->BuildQuery($recordType);

        eval { $queryDef->BuildField($dbidField); };
        if ($@) {
            &Logs("The record type $recordType is invalid");
            next;
        }
        $queryDef->BuildField($versionField);
        $queryDef->BuildField($lockOwnerField);
        $queryDef->BuildField($lockVersionField);

        # If lockOwnerField > 0, the record is locked, so filter lock owner
        $filterOp = $queryDef->BuildFilterOperator($CQPerlExt::CQ_BOOL_OP_AND);
        @unlockValue = ("0");
        $filterOp->BuildFilter($lockOwnerField, $CQPerlExt::CQ_COMP_OP_GT,
            \@unlockValue);
        $resultset = $cqSession->BuildResultSet($queryDef);

        # Execute the query
        $resultset->Execute();

        # Output scan information if log level is verbose
        if ($ARG{LOGLEVEL} =~ /verbose/i) {
            &Logs("Looking for $recordType record locks ...");
        }

        # Analyze locked records if necessary
        $locks = 0;
        $timeoutLocks = 0;
        $unlocks = 0;
        while ($resultset->MoveNext() == 1) {
            $dbid = $resultset->GetColumnValue(1);
            $version = $resultset->GetColumnValue(2);
            $lockOwner = $resultset->GetColumnValue(3);
            $lockVersion = $resultset->GetColumnValue(4);

            # Generate a unique hash key for locked record
            $lockKey = "$recordType $dbid $version $lockOwner $lockVersion";

            # Output the locked record information if log level is verbose
            if ($ARG{LOGLEVEL} =~ /verbose/i) {
                $logmsg = "\t$recordType ".DisplayName($recordType, $dbid).
                          ' is locked by ';

                if ($lockOwner ne '*') {
                    $logmsg .= "$lockOwner";
                }
                else {
                    $logmsg .= 'an unknown user';
                }
            }

            # Check whether the locked record was scanned last time
            if (exists($lastLockRecords{$lockKey})) {
                # Increase scan times
                $lockRecords{$lockKey} = $lastLockRecords{$lockKey} + 1;

                # Check the record if meeting time out
                if (($lastLockRecords{$lockKey} >= $timeout) &&
                    ($ARG{TIMEOUT} > 0)) {
                    # Output unlock information if log level is verbose
                    if ($ARG{LOGLEVEL} =~ /verbose/i) {
                        $logmsg .= ' and is timed out';
                    }

                    if ($ARG{UNLOCK} == 1) {
                        if ($ARG{LOGLEVEL} =~ /verbose/i) {
                            $logmsg .= '; it was unlocked';
                        }

                        # Call unlock subroutine
                        eval { UnlockRecord($recordType, $dbid); };
                        if ($@) {
                            &Logs($@);
                        }
                        else {
                            # Unlock successfully, clean up the lock information
                            delete($lockRecords{$lockKey});
                            delete($lastLockRecords{$lockKey});
                            $unlocks++;
                        }
                    }

                    # Count timeout locks
                    $timeoutLocks++;
                }
            }
            else {
                # Find the locked record first time
                $lockRecords{$lockKey} = 1;
            }

            # Count locks
            $locks++;

            if ($ARG{LOGLEVEL} =~ /verbose/i) {
                &Logs($logmsg);
            }
        }

        if ($ARG{LOGLEVEL} =~ /info|verbose/i) {
            &Logs("Summary for $recordType: $locks locks found, "
                  ."$timeoutLocks timed out, $unlocks unlocked");
        }

        # Count total locks
        $total += $locks;
        $totaltimeout += $timeoutLocks;
        $totalunlocks += $unlocks;
    }

    CQSession::Unbuild($cqSession);

    if ($ARG{LOGLEVEL} =~ /info|verbose/i) {
        &Logs("Scan summary: $total locks found, $totaltimeout timed out, "
              ."$totalunlocks unlocked");
        &Logs('');
    }

    # Backup latest lock records informaiton for time out if necessary later
    %lastLockRecords = %lockRecords;

    if (($ARG{SCAN} == 0) || ($scantimes == 1)) {
        # Report only
        last;
    }
    elsif ($scantimes > 0) {
        $scantimes--;
    }

    # Wait exactly for checking locks frequently
    $scanTime = $ARG{SCAN} - (time() - $beginTime);
    if ($scanTime > 0) {
        sleep($scanTime);
    }
}

# Get the display name associated with the Entity
sub DisplayName ($$) {
    my($recordType) = shift;
    my($dbid) = shift;
    my $record;
    eval { $record = $cqSession->GetEntityByDbId($recordType, $dbid); };
    if ($@) {
        return '';
    }
    else {
        return $record->GetDisplayName();
    }
}

# Unlock the record
sub UnlockRecord ($$) {
    my($recordType) = shift;
    my($dbid) = shift;
    my $record = $cqSession->GetEntityByDbId($recordType, $dbid);
    $record->UnlockRecord();
}

# Write log
sub Logs ($) {
    my $msg = shift;
    my @ltime = localtime;
    my $timestamp = sprintf("[%d-%02d-%02d %02d:%02d:%02d] ", (1900+$ltime[5]),
                    ($ltime[4]+1), $ltime[3], $ltime[2], $ltime[1], $ltime[0]);
    print $timestamp.$msg."\n";
}

undef $cqSession;
